/*
*	Life+ for Web - an interactive stochastic generalization of Conway's "Game of Life", for web pages
*	v0.9.3, 2016.06.29
*	Copyright (c) 2016 Sunkware
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*	See the GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program. If not, see <http://www.gnu.org/licenses/>.
*
*	WWW: sunkware.org				E-MAIL: sunkware@gmail.com
*/


// Primary parameters
var liferandFieldSizeLog = 7;
var liferandCellSize = 5;
var liferandFramerate = 25;

var liferandRainbowColors = new Array("#000000", "#d000f0", "#0000d0", "#0060ff", "#00e0e0", "#40c000", "#f0f000", "#f08000", "#f00000");

var liferandAboutString = "<p style=\"margin-top: 0px;\" title=\"Life+ for Web &mdash; an interactive stochastic generalization of Conway's Game of Life, for web pages\nv0.9.3, 2016.06.29\nCopyright &copy; 2016 Sunkware\nsunkware.org\"><b>&nbsp;&image;&nbsp;</b></p>";

var liferandTextColor = "#202000";
var liferandBackgroundColor = "#f0ffd8";

var liferandControlsWidth = 340;
var liferandControlsLeftPadding = 32;

var liferandControlsBaseFontSize = 18;
var liferandControlsHeaderFontSize = 22;
var liferandControlsSpeedFontSize = 20;
var liferandControlsInfoFontSize = 20;


// Derived parameters
var liferandFieldSize = 1 << liferandFieldSizeLog;
var liferandFieldSizeMask = liferandFieldSize - 1;

var liferandCanvasSize = liferandCellSize * liferandFieldSize;


// Well... palette
var liferandPalette = new Array(0x100 + 1);


// Control probabilities
var liferandProbs = new Array(2); // 0 --- probabilities to become alive for empty cell, 1 --- to stay alive for alive cell
liferandProbs[0] = new Array(0, 0, 0, 1.0, 0, 0, 0, 0, 0);
liferandProbs[1] = new Array(0, 0, 1.0, 1.0, 0, 0, 0, 0, 0);
// B3/S23


// Field data
var liferandFieldCurrent;
var liferandFieldNext;


var liferandPageLoaded = false;


// Sync timer id
var liferandSyncTimerID;
	

// Canvas, controls & other markup

document.write("<div style=\"float: left;\">");
document.write("<canvas id=\"liferand_canvas\" width=\"" + liferandCanvasSize + "\" height=\"" + liferandCanvasSize + "\"></canvas>");
document.write("</div>");


document.write("<div style=\"float: right; width: " + liferandControlsWidth + "px; padding-left: " + liferandControlsLeftPadding + "px; padding-bottom: 8px; color: " + liferandTextColor + "; background-color: " + liferandBackgroundColor + "; font-size: " + liferandControlsBaseFontSize + "px;\">");
document.write("<form>");

document.write("<p style=\"font-size: " + liferandControlsHeaderFontSize + "px; margin: 16px auto 8px auto;\">Empty &rarr; Alive:</p>");
for (var i = 0; i <= 8; i++) {
	document.write("<p style=\"margin: 0px auto 0px auto;\"><b>" + i + "</b>: &nbsp; <input type=\"range\" id=\"liferandSliderEmptyProbs" + i + "\" title=\"Probability for each empty cell with exactly " + i + " alive neighbour" + (i != 1 ? "s":"") + "\nto become alive\" min=\"0\" max=\"100\" step=\"0.5\" value=\"" + (100 * liferandProbs[0][i]) + "\" style=\"width: 200px;\" onChange=\"liferandEmptyProbsChange(" + i + ")\" /> &nbsp;&nbsp; <label id=\"liferandLabelEmptyProbs" + i + "\">" + (100 * liferandProbs[0][i]) + "</label></p>");
}

document.write("<p style=\"font-size: " + liferandControlsHeaderFontSize + "px; margin: 16px auto 8px auto;\">Alive &rarr; Alive:</p>");
for (var i = 0; i <= 8; i++) {
	document.write("<p style=\"margin: 0px auto 0px auto;\"><b>" + i + "</b>: &nbsp; <input type=\"range\" id=\"liferandSliderAliveProbs" + i + "\" title=\"Probability for each alive cell with exactly " + i + " alive neighbour" + (i != 1 ? "s":"") + "\nto stay alive\" min=\"0\" max=\"100\" step=\"0.5\" value=\"" + (100 * liferandProbs[1][i]) + "\" style=\"width: 200px;\" onChange=\"liferandAliveProbsChange(" + i + ")\" /> &nbsp;&nbsp; <label id=\"liferandLabelAliveProbs" + i + "\">" + (100 * liferandProbs[1][i]) + "</label></p>");
}

document.write("<hr style=\"margin-top: 12px;\">");


document.write("<table width=" + (liferandControlsWidth - liferandControlsLeftPadding) + "><tr><td>");

document.write("<p style=\"font-size: " + liferandControlsSpeedFontSize + "px; margin: 8px auto 0px auto;\"><b>Speed</b>: &nbsp; <input type=\"range\" id=\"liferandSliderFramerate\" title=\"Turns per second (roughly)\n0 to pause and cell-level edit\" min=\"0\" max=\"50\" step=\"1\" value=\"" + liferandFramerate + "\" style=\"width: 100px;\" onChange=\"liferandFramerateChange()\" /> &nbsp;&nbsp; <label id=\"liferandLabelFramerate\">" + liferandFramerate + "</label></p>");

document.write("</td><td style=\"text-align: right;\">");

document.write("<div style=\"width: 24px; height: 32px; font-size: " + liferandControlsInfoFontSize + "px; margin-top: 0px; color: #a0a0ff; background-color: #ffffa0; border: 1px solid #f0f080; border-radius: 12px;\">");
document.write(liferandAboutString);
document.write("</div>");

document.write("</td></tr></table>");


document.write("</form>");
document.write("</div>");


document.write("<div style=\"clear: both; height: 0px;\">&nbsp;</div>");

document.close();


var liferandCanvas = document.getElementById("liferand_canvas");
var liferandCanvasContext = liferandCanvas.getContext("2d");


// Handlers or probabilities changing by sliders
function liferandEmptyProbsChange(i) {
	if (liferandPageLoaded) {	
		var probSlider = document.getElementById("liferandSliderEmptyProbs" + i);
		var probLabel = document.getElementById("liferandLabelEmptyProbs" + i);

		liferandProbs[0][i] = probSlider.value / 100.0;
		probLabel.innerHTML = probSlider.value;
	}	
}

function liferandAliveProbsChange(i) {
	if (liferandPageLoaded) {
		var probSlider = document.getElementById("liferandSliderAliveProbs" + i);
		var probLabel = document.getElementById("liferandLabelAliveProbs" + i);

		liferandProbs[1][i] = probSlider.value / 100.0;
		probLabel.innerHTML = probSlider.value;
	}
}

// Handler of framerate changing
function liferandFramerateChange() {
	if (liferandPageLoaded) {
		liferandFramerate = liferandSliderFramerate.value;
		liferandLabelFramerate.innerHTML = liferandFramerate;

		if (liferandSyncTimerID >= 0) {
			window.clearInterval(liferandSyncTimerID);
			liferandSyncTimerID = -1;
		}

		if (liferandFramerate > 0) {
			liferandSyncTimerID = window.setInterval("liferandSyncField()", 1000 / liferandFramerate);
		}
	}
}


// Additional low-level math...
function liferandHeaviside(value) { // Should be 1 for positive value, 0 for the rest
	return (value > 0) ? 1 : 0; // exercise: optimize it platform-independently to avoid execution branches... in the absence of Math.sign()
}

function liferandIncBinLog(value, maxLog) { // 0 for 0, then incremented binary logarithm up to maxLog
	var res = 0;

	while ((value > 0) && (res < maxLog)) {
		value >>= 1;
		res++;
	}

	return res;
}


// Field-related functions...
function liferandInitField() {
	liferandFieldCurrent = new Array(liferandFieldSize);
	liferandFieldNext = new Array(liferandFieldSize);

	for (var y = 0; y < liferandFieldSize; y++) {
		liferandFieldCurrent[y] = new Array(liferandFieldSize);
		liferandFieldNext[y] = new Array(liferandFieldSize);
	}
}

function liferandClearField() {
	for (var y = 0; y < liferandFieldSize; y++) {
		for (var x = 0; x < liferandFieldSize; x++) {
			liferandFieldCurrent[y][x] = 0;
		}
	}
}

function liferandInhabitField() {
	for (var y = 0; y < liferandFieldSize; y++) {
		for (var x = 0; x < liferandFieldSize; x++) {
			liferandFieldCurrent[y][x] = Math.floor(2 * Math.random());
		}
	}
}

function liferandFlipCurrentNext() { // swapping pointers instead of copying...
	var storedPointer = liferandFieldCurrent;

	liferandFieldCurrent = liferandFieldNext;
	liferandFieldNext = storedPointer;
}

function liferandGetNeighbours(y, x) {
	var res = 0;

	res += liferandHeaviside(liferandFieldCurrent[y][(x + 1) & liferandFieldSizeMask]);
	res += liferandHeaviside(liferandFieldCurrent[(y + 1) & liferandFieldSizeMask][(x + 1) & liferandFieldSizeMask]);
	res += liferandHeaviside(liferandFieldCurrent[(y + 1) & liferandFieldSizeMask][x]);
	res += liferandHeaviside(liferandFieldCurrent[(y + 1) & liferandFieldSizeMask][(x - 1) & liferandFieldSizeMask]);
	res += liferandHeaviside(liferandFieldCurrent[y][(x - 1) & liferandFieldSizeMask]);
	res += liferandHeaviside(liferandFieldCurrent[(y - 1) & liferandFieldSizeMask][(x - 1) & liferandFieldSizeMask]);
	res += liferandHeaviside(liferandFieldCurrent[(y - 1) & liferandFieldSizeMask][x]);
	res += liferandHeaviside(liferandFieldCurrent[(y - 1) & liferandFieldSizeMask][(x + 1) & liferandFieldSizeMask]);

	return res;
}

function liferandUpdateField() { // main one...

	for (var y = 0; y < liferandFieldSize; y++) {
		for (var x = 0; x < liferandFieldSize; x++) {

			var cellAge = liferandFieldCurrent[y][x];
			var cellState = liferandHeaviside(cellAge);

			var cellNeighbours = liferandGetNeighbours(y, x);

			var cellNextState = liferandHeaviside(liferandProbs[cellState][cellNeighbours] - Math.random());
				
			var deltaAge = cellNextState + (1 - cellNextState) * (-cellAge * cellState);
// if cellNextState = 1, it's 1 regardless of current state (age increments);
// otherwise, it's 0 for empty cell and (-cellAge) for alive one
// (so that cellAge + (-cellAge) = 0 => cell becomes empty)

			liferandFieldNext[y][x] = cellAge + deltaAge;
		}
	}

	liferandFlipCurrentNext();
}


// Graphics...
function liferandDrawRect(x, y, w, h, color) {
	liferandCanvasContext.fillStyle = color;
	liferandCanvasContext.fillRect(x, y, w, h);
}

function liferandDrawField() {
	for (var y = 0; y < liferandFieldSize; y++) {
		for (var x = 0 ; x < liferandFieldSize; x++) {
			var colorIndex = Math.min(0x100, liferandFieldCurrent[y][x]);

			liferandDrawRect(x * liferandCellSize, y * liferandCellSize, liferandCellSize, liferandCellSize, liferandPalette[colorIndex]);
		}
	}
}


function liferandSyncField() {
	liferandUpdateField();
	liferandDrawField();
}


// Generating palette...
function liferandMakePalette_RainbowLog() {
	liferandPalette[0] = liferandRainbowColors[0];

	for (var i = 1; i < 0x100; i++) {
		liferandPalette[i] = liferandRainbowColors[liferandIncBinLog(i, 8)];
	}

	liferandPalette[0x100] = liferandRainbowColors[liferandRainbowColors.length - 1];
}


// Initialization...
window.onload = function() {
	liferandPageLoaded = true;

	liferandInitField();

	liferandInhabitField();

	liferandMakePalette_RainbowLog();

	liferandSyncTimerID = window.setInterval("liferandSyncField()", 1000 / liferandFramerate);
}


// Edit field by click on it...
liferandCanvas.onclick = function(event) {
	if (liferandPageLoaded) {

		if (liferandFramerate == 0) {
			var fieldX = Math.floor(event.offsetX / liferandCellSize);
			var fieldY = Math.floor(event.offsetY / liferandCellSize);

			if ((fieldX >= 0) && (fieldX < liferandFieldSize) && (fieldY >= 0) && (fieldY < liferandFieldSize)) {
				liferandFieldCurrent[fieldY][fieldX] = 1 - liferandHeaviside(liferandFieldCurrent[fieldY][fieldX]); // empty becomes alive, and vice versa
			}

		} else {
			liferandClearField();
		}

		liferandDrawField();
	}
}

liferandCanvas.ondblclick = function() {
	if (liferandPageLoaded) {
		if (liferandFramerate > 0) {
			liferandInhabitField();
			liferandDrawField();
		}
	}
}





